@@ -10,6 +10,7 @@ import os.path
 import posixpath
 import queue
 import re
+import sys
 import threading
 import time
 from collections import namedtuple
@@ -103,26 +104,50 @@ class Repository:
     # Fast KDF for high-entropy inputs (used for shared data)
     DEFAULT_SHARED_KDF_NAME = 'blake2b'
     EMPTY_TABLE_VALUE = '--'
+    DEFAULT_CACHE_DIRECTORY = utils.fs.DEFAULT_CACHE_DIRECTORY
 
-    def __init__(self, backend, *, concurrent, quiet=True):
+    def __init__(
+        self,
+        backend,
+        *,
+        concurrent,
+        quiet=True,
+        cache_directory=DEFAULT_CACHE_DIRECTORY,
+    ):
         self._concurrent = concurrent
         self._quiet = quiet
+        self._cache_directory = cache_directory
+
         self._slots = asyncio.Queue(maxsize=concurrent)
         # We need actual integers for TQDM slot management in CLI, but this queue
-        # can also act as a semaphore in the general case
+        # can also act as a semaphore in the general case. Note that numbering starts
+        # from 2, leaving two topmost slots to non-slotable trackers
         for slot in range(2, concurrent + 2):
             self._slots.put_nowait(slot)
 
         self.backend = backend
 
+    def display_status(self, message):
+        print(ef.bold + message + ef.rs, file=sys.stderr)
+
+    def _get_cached(self, path):
+        assert self._cache_directory is not None
+        return Path(self._cache_directory, path).read_bytes()
+
+    def _store_cached(self, path, data):
+        assert self._cache_directory is not None
+        file = Path(self._cache_directory, path)
+        file.parent.mkdir(parents=True, exist_ok=True)
+        file.write_bytes(data)
+
     @property
+    def _unlocked(self):
+        return hasattr(self, 'props')
+
+    @cached_property
     def executor(self):
         """Executor for non-async methods of the backend instance"""
-        try:
-            return self._executor
-        except AttributeError:
-            self._executor = ThreadPoolExecutor(max_workers=self._concurrent)
-            return self._executor
+        return ThreadPoolExecutor(max_workers=self._concurrent)
 
     def _awrap(self, func, *args, **kwargs):
         if inspect.iscoroutinefunction(func) or inspect.isasyncgenfunction(func):
@@ -144,10 +169,6 @@ class Repository:
         async for value in result:
             yield value
 
-    @property
-    def _unlocked(self):
-        return hasattr(self, 'props')
-
     @asynccontextmanager
     async def _acquire_slot(self):
         slot = await self._slots.get()
@@ -156,6 +177,33 @@ class Repository:
         finally:
             self._slots.put_nowait(slot)
 
+    # Convenience wrappers for backend methods that do not need to know
+    # the actual slot they are occupying (so most of them)
+    async def _exists(self, location):
+        async with self._acquire_slot():
+            logger.info('Checking existence of %s', location)
+            return await self._awrap(self.backend.exists, location)
+
+    async def _download(self, location):
+        async with self._acquire_slot():
+            logger.info('Downloading %s', location)
+            return await self._awrap(self.backend.download, location)
+
+    async def _upload_data(self, location, data):
+        async with self._acquire_slot():
+            logger.info('Uploading binary data to %s', location)
+            await self._awrap(self.backend.upload, location, data)
+
+    async def _delete(self, location):
+        async with self._acquire_slot():
+            logger.info('Deleting %s', location)
+            await self._awrap(self.backend.delete, location)
+
+    async def _clean(self):
+        async with self._acquire_slot():
+            logger.info('Running backend clean-up')
+            await self._awrap(self.backend.clean)
+
     def default_serialization_hook(self, data, /):
         return utils.type_hint(data)
 
@@ -244,6 +292,29 @@ class Repository:
             'st_ctime': stat_result.st_ctime,
         }
 
+    def _validate_settings(self, schema, obj):
+        extra_keys = obj.keys() - schema.keys()
+        if extra_keys:
+            raise exceptions.ReplicatError(
+                'Settings include unrecognized properties: '
+                + ', '.join(map(repr, extra_keys))
+            )
+
+        for key, value in obj.items():
+            allowed_types = schema[key]
+            if isinstance(value, allowed_types):
+                continue
+
+            if isinstance(allowed_types, tuple):
+                type_names = ', '.join([x.__name__ for x in allowed_types])
+            else:
+                type_names = allowed_types.__name__
+
+            raise exceptions.ReplicatError(
+                f'{key!r} -> {value!r} does not match any '
+                f'of the allowed types ({type_names})'
+            )
+
     def _make_config(self, *, settings=None):
         if settings is None:
             settings = {}
@@ -360,15 +431,37 @@ class Repository:
             'private': private,
         }
 
+    def _validate_init_settings(self, settings):
+        self._validate_settings(
+            {
+                'hashing': collections.abc.Mapping,
+                'chunking': collections.abc.Mapping,
+                'encryption': (collections.abc.Mapping, type(None)),
+            },
+            settings,
+        )
+
+        if (encryption_settings := settings.get('encryption')) is not None:
+            self._validate_settings(
+                {
+                    'cipher': collections.abc.Mapping,
+                    'kdf': collections.abc.Mapping,
+                },
+                encryption_settings,
+            )
+
     async def init(self, *, password=None, settings=None, key_output_path=None):
+        if settings:
+            self._validate_init_settings(settings)
+
         logger.info('Using provided settings: %r', settings)
-        print(ef.bold + 'Generating new config' + ef.rs)
+        self.display_status('Generating new config')
         config = self._make_config(settings=settings)
         print(json.dumps(config, indent=4, default=self.default_serialization_hook))
         props = RepositoryProps(**self._instantiate_config(config))
 
         if props.encrypted:
-            print(ef.bold + 'Generating new key' + ef.rs)
+            self.display_status('Generating new key')
             if password is None:
                 raise exceptions.ReplicatError(
                     'A password is needed to initialize encrypted repository'
@@ -390,7 +483,7 @@ class Repository:
             # TODO: store keys in the repository?
             if key_output_path is not None:
                 key_output_path = Path(key_output_path).resolve()
-                print(ef.bold + f'Writing key to {key_output_path}' + ef.rs)
+                self.display_status(f'Writing key to {key_output_path}')
                 key_output_path.write_bytes(self.serialize(key))
             else:
                 print(
@@ -399,25 +492,22 @@ class Repository:
         else:
             key = None
 
-        async with self._acquire_slot():
-            print(ef.bold + 'Uploading config' + ef.rs)
-            await self._awrap(self.backend.upload, 'config', self.serialize(config))
+        self.display_status('Uploading config')
+        await self._upload_data('config', self.serialize(config))
 
         self.props = props
         return utils.DefaultNamespace(config=config, key=key)
 
     async def _load_config(self):
-        async with self._acquire_slot():
-            data = await self._awrap(self.backend.download, 'config')
-
+        data = await self._download('config')
         return RepositoryProps(**self._instantiate_config(self.deserialize(data)))
 
     async def unlock(self, *, password=None, key=None):
-        print(ef.bold + 'Loading config' + ef.rs)
+        self.display_status('Loading config')
         props = await self._load_config()
 
         if props.encrypted:
-            print(ef.bold + 'Unlocking repository' + ef.rs)
+            self.display_status('Unlocking repository')
             if password is None or key is None:
                 raise exceptions.ReplicatError(
                     'Both password and key are needed to unlock this repository'
@@ -435,7 +525,7 @@ class Repository:
         self.props = props
 
     async def _add_key(self, *, password, settings, props, private, key_output_path):
-        print(ef.bold + 'Generating new key' + ef.rs)
+        self.display_status('Generating new key')
         key = self._make_key(
             cipher=props.cipher,
             chunker=props.chunker,
@@ -453,16 +543,29 @@ class Repository:
         # TODO: store it in the repository?
         if key_output_path is not None:
             key_output_path = Path(key_output_path).resolve()
-            print(ef.bold + f'Writing key to {key_output_path}' + ef.rs)
+            self.display_status(f'Writing key to {key_output_path}')
             key_output_path.write_bytes(self.serialize(key))
         else:
             print(json.dumps(key, indent=4, default=self.default_serialization_hook))
 
         return key
 
+    def _validate_add_key_settings(self, settings):
+        self._validate_settings(
+            {'encryption': collections.abc.Mapping},
+            settings,
+        )
+        self._validate_settings(
+            {'kdf': collections.abc.Mapping},
+            settings['encryption'],
+        )
+
     async def add_key(
         self, *, password, settings=None, shared=False, key_output_path=None
     ):
+        if settings:
+            self._validate_add_key_settings(settings)
+
         logger.info('Using provided settings: %r', settings)
         if password is None:
             raise exceptions.ReplicatError(
@@ -479,7 +582,7 @@ class Repository:
             if self._unlocked:
                 props = self.props
             else:
-                print(ef.bold + 'Loading config' + ef.rs)
+                self.display_status('Loading config')
                 props = await self._load_config()
 
         if not props.encrypted:
@@ -494,26 +597,52 @@ class Repository:
         )
         return utils.DefaultNamespace(new_key=key)
 
-    async def _load_snapshots(self, *, snapshot_regex=None, cache_loaded=True):
-        snapshots = {}
-        tasks = []
-        cached_snapshots = {str(x) for x in utils.fs.list_cached(self.SNAPSHOT_PREFIX)}
-        encrypted_snapshots = {}
+    async def _load_snapshots(self, *, snapshot_regex=None):
+        empty = object()
+        task_to_path = {}
 
         async def _download_snapshot(path, digest):
-            async with self._acquire_slot():
-                logger.info('Downloading %s', path)
-                contents = await self._awrap(self.backend.download, path)
+            contents = empty
+            if self._cache_directory is not None:
+                try:
+                    contents = self._get_cached(path)
+                except FileNotFoundError:
+                    pass
 
-            logger.info('Verifying %s', path)
-            if self.props.hash_digest(contents) != digest:
-                raise exceptions.ReplicatError(f'Snapshot at {path!r} is corrupted')
+            if contents is empty:
+                contents = await self._download(path)
+                logger.info('Verifying %s', path)
+                if self.props.hash_digest(contents) != digest:
+                    raise exceptions.ReplicatError(f'Snapshot at {path!r} is corrupted')
 
-            if cache_loaded:
-                logger.info('Caching %s', path)
-                utils.fs.store_cached(path, contents)
+                if self._cache_directory is not None:
+                    logger.info('Caching %s', path)
+                    self._store_cached(path, contents)
 
-            encrypted_snapshots[path] = contents
+            body = self.deserialize(contents)
+
+            if self.props.encrypted:
+                logger.info('Decrypting %s', path)
+                body['chunks'] = self.deserialize(
+                    self.props.decrypt(
+                        body['chunks'],
+                        self.props.derive_shared_subkey(
+                            self.props.hash_digest(body['data'])
+                        ),
+                    )
+                )
+                try:
+                    data = self.props.decrypt(body['data'], self.props.userkey)
+                except exceptions.DecryptionError:
+                    logger.info(
+                        "Decryption of %s failed, but it's not corrupted (different key?)",
+                        path,
+                    )
+                    body['data'] = None
+                else:
+                    body['data'] = self.deserialize(data)
+
+            return body
 
         async for path in self._aiter(self.backend.list_files, self.SNAPSHOT_PREFIX):
             name, tag = self.parse_snapshot_location(path)
@@ -527,41 +656,11 @@ class Repository:
                 logger.info('Skipping %s (does not match the filter)', path)
                 continue
 
-            if path in cached_snapshots:
-                encrypted_snapshots[path] = utils.fs.get_cached(path)
-            else:
-                tasks.append(asyncio.create_task(_download_snapshot(path, digest)))
+            task = asyncio.create_task(_download_snapshot(path, digest))
+            task_to_path[task] = path
 
-        if tasks:
-            await asyncio.gather(*tasks)
-
-        for path, contents in encrypted_snapshots.items():
-            snapshot_body = self.deserialize(contents)
-
-            if self.props.encrypted:
-                logger.info('Decrypting %s', path)
-                snapshot_body['chunks'] = self.deserialize(
-                    self.props.decrypt(
-                        snapshot_body['chunks'],
-                        self.props.derive_shared_subkey(
-                            self.props.hash_digest(snapshot_body['data'])
-                        ),
-                    )
-                )
-                try:
-                    data = self.props.decrypt(snapshot_body['data'], self.props.userkey)
-                except exceptions.DecryptionError:
-                    logger.info(
-                        "Decryption of %s failed, but it's not corrupted (different key?)",
-                        path,
-                    )
-                    snapshot_body['data'] = None
-                else:
-                    snapshot_body['data'] = self.deserialize(data)
-
-            snapshots[path] = snapshot_body
-
-        return snapshots
+        async for task in utils.as_completed(task_to_path):
+            yield task_to_path[task], await task
 
     def _format_snapshot_name(self, *, path, chunks, data):
         return self.parse_snapshot_location(path).name
@@ -590,11 +689,6 @@ class Repository:
         return utils.bytes_to_human(sum(r[1] - r[0] for r in ranges_it))
 
     async def list_snapshots(self, *, snapshot_regex=None):
-        print(ef.bold + 'Loading snapshots' + ef.rs)
-        snapshots_mapping = await self._load_snapshots(snapshot_regex=snapshot_regex)
-        if not snapshots_mapping:
-            return
-
         columns_getters = {
             'snapshot': self._format_snapshot_name,
             'note': self._format_snapshot_note,
@@ -603,26 +697,37 @@ class Repository:
             'size': self._format_snaphot_size,
         }
         columns_widths = {}
-        sorted_snapshots = sorted(
-            snapshots_mapping.items(),
-            key=lambda x: (x[1]['data'] or {}).get('utc_timestamp', ''),
-            reverse=True,
-        )
-        rows = []
+        snapshots = []
 
-        for snapshot_path, snapshot_body in sorted_snapshots:
+        self.display_status('Loading snapshots')
+        async for snapshot_path, snapshot_body in self._load_snapshots(
+            snapshot_regex=snapshot_regex
+        ):
             row = {}
+            snapshot_chunks = snapshot_body['chunks']
+            snapshot_data = snapshot_body['data']
+
             for column, getter in columns_getters.items():
                 value = getter(
                     path=snapshot_path,
-                    chunks=snapshot_body['chunks'],
-                    data=snapshot_body['data'],
+                    chunks=snapshot_chunks,
+                    data=snapshot_data,
                 )
                 value = str(value if value is not None else self.EMPTY_TABLE_VALUE)
                 row[column] = value
                 columns_widths[column] = max(len(value), columns_widths.get(column, 0))
 
-            rows.append(row)
+            if snapshot_data is not None:
+                snapshot_timestamp = snapshot_data['utc_timestamp']
+            else:
+                snapshot_timestamp = None
+
+            snapshots.append((snapshot_timestamp, row))
+
+        if not snapshots:
+            return
+
+        snapshots.sort(key=lambda x: x[0] or '', reverse=True)
 
         formatted_headers = []
         for column in columns_widths:
@@ -630,7 +735,7 @@ class Repository:
             formatted_headers.append(column.upper().center(width))
 
         print(*formatted_headers)
-        for row in rows:
+        for _, row in snapshots:
             print(*(value.ljust(columns_widths[col]) for col, value in row.items()))
 
     def _format_file_snapshot_date(
@@ -656,11 +761,6 @@ class Repository:
         return utils.bytes_to_human(sum(r[1] - r[0] for r in ranges_it))
 
     async def list_files(self, *, snapshot_regex=None, files_regex=None):
-        print(ef.bold + 'Loading snapshots' + ef.rs)
-        snapshots_mapping = await self._load_snapshots(snapshot_regex=snapshot_regex)
-        if not snapshots_mapping:
-            return
-
         columns_getters = {
             'snapshot date': self._format_file_snapshot_date,
             'path': self._format_file_path,
@@ -670,7 +770,10 @@ class Repository:
         columns_widths = {}
         files = []
 
-        for snapshot_path, snapshot_body in snapshots_mapping.items():
+        self.display_status('Loading snapshots')
+        async for snapshot_path, snapshot_body in self._load_snapshots(
+            snapshot_regex=snapshot_regex
+        ):
             snapshot_chunks = snapshot_body['chunks']
             snapshot_data = snapshot_body['data']
             if snapshot_data is None:
@@ -715,7 +818,7 @@ class Repository:
         return list(utils.fs.flatten_paths(path.resolve(strict=True) for path in paths))
 
     async def snapshot(self, *, paths, note=None, rate_limit=None):
-        print(ef.bold + 'Collecting files' + ef.rs)
+        self.display_status('Collecting files')
         files = self._flatten_paths(paths)
         logger.info('Found %d files', len(files))
         # Small files are more likely to change than big files, read them quickly
@@ -879,9 +982,7 @@ class Repository:
                     await asyncio.sleep(queue_timeout)
                     continue
 
-                async with self._acquire_slot():
-                    exists = await self._awrap(self.backend.exists, chunk.location)
-
+                exists = await self._exists(chunk.location)
                 logger.info(
                     'Processing chunk #%d, exists=%s, I=%d, L=%s',
                     chunk.counter,
@@ -918,10 +1019,10 @@ class Repository:
             ThreadPoolExecutor(max_workers=1, thread_name_prefix='chunk-producer'),
             _chunk_producer,
         )
-        with bytes_tracker, finished_tracker:
+        with finished_tracker, bytes_tracker:
             try:
                 await asyncio.gather(*(_worker() for _ in range(self._concurrent)))
-            except BaseException:
+            except:
                 abort.set()
                 raise
             finally:
@@ -967,16 +1068,12 @@ class Repository:
         name, tag = self._snapshot_digest_to_location_parts(digest)
         location = self.get_snapshot_location(name=name, tag=tag)
 
-        async with self._acquire_slot():
-            print(ef.bold + f'Uploading snapshot {name}' + ef.rs)
-            await self._awrap(self.backend.upload, location, serialized_snapshot)
+        self.display_status(f'Uploading snapshot {name}')
+        await self._upload_data(location, serialized_snapshot)
 
         if state.bytes_reused:
-            print(
-                ef.bold
-                + f'Used {utils.bytes_to_human(state.bytes_reused)} of existing data'
-                + ef.rs,
-            )
+            reused_human = utils.bytes_to_human(state.bytes_reused)
+            self.display_status(f'Used {reused_human} of existing data')
 
         return utils.DefaultNamespace(
             name=name,
@@ -994,9 +1091,14 @@ class Repository:
 
         logger.info("Will restore files to %s", path)
 
-        print(ef.bold + 'Loading snapshots' + ef.rs)
-        snapshots_mapping = await self._load_snapshots(snapshot_regex=snapshot_regex)
-        snapshots = [x for x in snapshots_mapping.values() if x['data'] is not None]
+        self.display_status('Loading snapshots')
+        snapshots = []
+        async for _, snapshot_body in self._load_snapshots(
+            snapshot_regex=snapshot_regex
+        ):
+            if snapshot_body['data'] is not None:
+                snapshots.append(snapshot_body)
+
         snapshots.sort(key=lambda x: x['data']['utc_timestamp'], reverse=True)
 
         chunks_references = {}
@@ -1053,10 +1155,7 @@ class Repository:
                     break
 
                 location = self._chunk_digest_to_location(digest)
-
-                async with self._acquire_slot():
-                    logger.info('Downloading %s', location)
-                    contents = await self._awrap(self.backend.download, location)
+                contents = await self._download(location)
 
                 logger.info('Decrypting %s', location)
                 if self.props.encrypted:
@@ -1138,33 +1237,32 @@ class Repository:
 
     async def delete_snapshots(self, snapshots):
         # TODO: locking
-        print(ef.bold + 'Loading snapshots' + ef.rs)
-        snapshots_mapping = await self._load_snapshots()
-        chunks_digests = set()
+        self.display_status('Loading snapshots')
+        to_delete = set()
+        to_keep = set()
         snapshots_locations = set()
         remaining_names = set(snapshots)
 
-        for location in list(snapshots_mapping):
-            name = self.parse_snapshot_location(location).name
+        async for path, body in self._load_snapshots():
+            name = self.parse_snapshot_location(path).name
             if name in remaining_names:
-                body = snapshots_mapping[location]
                 if body['data'] is None:
                     raise exceptions.ReplicatError(
                         f'Cannot delete snapshot {name} (different key)'
                     )
 
-                chunks_digests.update(body['chunks'])
-                snapshots_locations.add(location)
+                to_delete.update(body['chunks'])
+                snapshots_locations.add(path)
                 remaining_names.discard(name)
+            else:
+                to_keep.update(body['chunks'])
 
         if remaining_names:
             raise exceptions.ReplicatError(
                 f'Snapshots {", ".join(remaining_names)} are not available'
             )
 
-        for location, body in snapshots_mapping.items():
-            if location not in snapshots_locations:
-                chunks_digests.difference_update(body['chunks'])
+        to_delete.difference_update(to_keep)
 
         finished_snapshots_tracker = tqdm(
             desc='Snapshots deleted',
@@ -1176,8 +1274,7 @@ class Repository:
         )
 
         async def _delete_snapshot(location):
-            async with self._acquire_slot():
-                await self._awrap(self.backend.delete, location)
+            await self._delete(location)
             finished_snapshots_tracker.update()
 
         with finished_snapshots_tracker:
@@ -1186,35 +1283,32 @@ class Repository:
         finished_chunks_tracker = tqdm(
             desc='Unreferenced chunks deleted',
             unit='',
-            total=len(chunks_digests),
+            total=len(to_delete),
             position=0,
             disable=self._quiet,
             leave=True,
         )
 
         async def _delete_chunk(digest):
-            async with self._acquire_slot():
-                await self._awrap(
-                    self.backend.delete, self._chunk_digest_to_location(digest)
-                )
+            location = self._chunk_digest_to_location(digest)
+            await self._delete(location)
             finished_chunks_tracker.update()
 
         with finished_chunks_tracker:
-            await asyncio.gather(*map(_delete_chunk, chunks_digests))
+            await asyncio.gather(*map(_delete_chunk, to_delete))
 
     async def clean(self):
         # TODO: locking
-        print(ef.bold + 'Loading snapshots' + ef.rs)
-        snapshots_mapping = await self._load_snapshots()
+        self.display_status('Loading snapshots')
         referenced_digests = {
-            y for x in snapshots_mapping.values() for y in x['chunks']
+            y async for _, x in self._load_snapshots() for y in x['chunks']
         }
         referenced_locations = set(
             map(self._chunk_digest_to_location, referenced_digests)
         )
         to_delete = set()
 
-        print(ef.bold + 'Fetching chunks list' + ef.rs)
+        self.display_status('Fetching chunks list')
         async for location in self._aiter(self.backend.list_files, self.CHUNK_PREFIX):
             if location in referenced_locations:
                 logger.info('Chunk %s is referenced, skipping', location)
@@ -1230,7 +1324,6 @@ class Repository:
             to_delete.add(location)
 
         if not to_delete:
-            print('Nothing to do')
             return
 
         finished_tracker = tqdm(
@@ -1243,16 +1336,14 @@ class Repository:
         )
 
         async def _delete_chunk(location):
-            async with self._acquire_slot():
-                await self._awrap(self.backend.delete, location)
+            await self._delete(location)
             finished_tracker.update()
 
         with finished_tracker:
             await asyncio.gather(*map(_delete_chunk, to_delete))
 
-        async with self._acquire_slot():
-            print(ef.bold + 'Running post-deletion cleanup' + ef.rs)
-            await self._awrap(self.backend.clean)
+        self.display_status('Running post-deletion cleanup')
+        await self._clean()
 
     @utils.disable_gc
     def _benchmark_chunker(self, adapter, number=10, size=512_000_000):
@@ -1275,9 +1366,9 @@ class Repository:
         processed_bytes = sum(map(len, adapter(stream)))
         elapsed = Decimal(time.perf_counter_ns() - start - prep_time).scaleb(-9)
         rate = processed_bytes / elapsed
-        print(
-            ef.bold + f'Processed {utils.bytes_to_human(processed_bytes, 3)} '
-            f'in {elapsed:.3f} seconds ({utils.bytes_to_human(rate, 3)}/s)' + ef.rs
+        self.display_status(
+            f'Processed {utils.bytes_to_human(processed_bytes, 3)} '
+            f'in {elapsed:.3f} seconds ({utils.bytes_to_human(rate, 3)}/s)'
         )
 
     async def benchmark(self, name, settings=None):
@@ -1290,15 +1381,18 @@ class Repository:
         argument_string = ', '.join(
             f'{name}={value!r}' for name, value in adapter_args.items()
         )
-        print(ef.bold + f'Benchmarking {name}({argument_string})' + ef.rs)
+        self.display_status(f'Benchmarking {name}({argument_string})')
 
         if isinstance(adapter, ChunkerAdapter):
-            self._benchmark_chunker(adapter)
+            loop = asyncio.get_running_loop()
+            await loop.run_in_executor(
+                ThreadPoolExecutor(max_workers=1), self._benchmark_chunker, adapter
+            )
         else:
             raise RuntimeError('Sorry, not yet')
 
     async def upload(self, paths, *, rate_limit=None, skip_existing=False):
-        print(ef.bold + 'Collecting files' + ef.rs)
+        self.display_status('Collecting files')
         files = self._flatten_paths(paths)
         logger.info('Found %d files', len(files))
         bytes_tracker = tqdm(
@@ -1330,13 +1424,8 @@ class Repository:
                 os.path.commonpath([path, base_directory])
             ).as_posix()
             length = path.stat().st_size
-            exists = False
 
-            if skip_existing:
-                async with self._acquire_slot():
-                    exists = await self._awrap(self.backend.exists, name)
-
-            if exists:
+            if skip_existing and await self._exists(name):
                 logger.info('Skipping file %r (exists)', name)
             else:
                 async with self._acquire_slot() as slot:
