@@ -4,7 +4,7 @@ from pathlib import Path
 
 from appdirs import user_cache_dir
 
-CACHE_DIRECTORY = user_cache_dir('replicat', 'replicat')
+DEFAULT_CACHE_DIRECTORY = user_cache_dir('replicat', 'replicat')
 DEFAULT_STREAM_CHUNK_SIZE = 16_777_216
 
 
@@ -39,21 +39,3 @@ def stream_files(files, *, chunk_size=DEFAULT_STREAM_CHUNK_SIZE):
                 yield path, chunk
                 if len(chunk) < chunk_size:
                     break
-
-
-def list_cached(path):
-    start = Path(CACHE_DIRECTORY, path)
-    start.mkdir(parents=True, exist_ok=True)
-
-    for path in flatten_paths([start]):
-        yield path.relative_to(CACHE_DIRECTORY)
-
-
-def get_cached(path):
-    return Path(CACHE_DIRECTORY, path).read_bytes()
-
-
-def store_cached(path, data):
-    file = Path(CACHE_DIRECTORY, path)
-    file.parent.mkdir(parents=True, exist_ok=True)
-    file.write_bytes(data)
